/*
 * This is a machine-generated file. Please do not edit it by hand.
 * (As of sbcl-0.8.14, it came from WRITE-CONFIG-H in genesis.lisp.)
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_CONFIG
#define SBCL_GENESIS_CONFIG
#define LISP_FEATURE_64_BIT
#define LISP_FEATURE_64_BIT_REGISTERS
#define LISP_FEATURE_ALIEN_CALLBACKS
#define LISP_FEATURE_ANSI_CL
#define LISP_FEATURE_ASH_RIGHT_VOPS
#define LISP_FEATURE_BSD
#define LISP_FEATURE_COMMON_LISP
#define LISP_FEATURE_COMPACT_INSTANCE_HEADER
#define LISP_FEATURE_COMPARE_AND_SWAP_VOPS
#define LISP_FEATURE_COMPLEX_FLOAT_VOPS
#define LISP_FEATURE_CYCLE_COUNTER
#define LISP_FEATURE_C_STACK_IS_CONTROL_STACK
#define LISP_FEATURE_DARWIN
#define LISP_FEATURE_DARWIN9_OR_BETTER
#define LISP_FEATURE_FLOAT_EQL_VOPS
#define LISP_FEATURE_FP_AND_PC_STANDARD_SAVE
#define LISP_FEATURE_GENCGC
#define LISP_FEATURE_IEEE_FLOATING_POINT
#define LISP_FEATURE_IMMOBILE_CODE
#define LISP_FEATURE_IMMOBILE_SPACE
#define LISP_FEATURE_INLINE_CONSTANTS
#define LISP_FEATURE_INODE64
#define LISP_FEATURE_INTEGER_EQL_VOP
#define LISP_FEATURE_LINKAGE_TABLE
#define LISP_FEATURE_LITTLE_ENDIAN
#define LISP_FEATURE_MACH_EXCEPTION_HANDLER
#define LISP_FEATURE_MACH_O
#define LISP_FEATURE_MEMORY_BARRIER_VOPS
#define LISP_FEATURE_MULTIPLY_HIGH_VOPS
#define LISP_FEATURE_OS_PROVIDES_BLKSIZE_T
#define LISP_FEATURE_OS_PROVIDES_DLADDR
#define LISP_FEATURE_OS_PROVIDES_DLOPEN
#define LISP_FEATURE_OS_PROVIDES_PUTWC
#define LISP_FEATURE_OS_PROVIDES_SUSECONDS_T
#define LISP_FEATURE_PACKAGE_LOCAL_NICKNAMES
#define LISP_FEATURE_PRECISE_ARG_COUNT_ERROR
#define LISP_FEATURE_RAW_INSTANCE_INIT_VOPS
#define LISP_FEATURE_RAW_SIGNED_WORD
#define LISP_FEATURE_RELOCATABLE_HEAP
#define LISP_FEATURE_SBCL
#define LISP_FEATURE_SB_CORE_COMPRESSION
#define LISP_FEATURE_SB_DOC
#define LISP_FEATURE_SB_EVAL
#define LISP_FEATURE_SB_LDB
#define LISP_FEATURE_SB_PACKAGE_LOCKS
#define LISP_FEATURE_SB_SIMD_PACK
#define LISP_FEATURE_SB_SOURCE_LOCATIONS
#define LISP_FEATURE_SB_THREAD
#define LISP_FEATURE_SB_TRACEROOT
#define LISP_FEATURE_SB_UNICODE
#define LISP_FEATURE_STACK_ALLOCATABLE_CLOSURES
#define LISP_FEATURE_STACK_ALLOCATABLE_FIXED_OBJECTS
#define LISP_FEATURE_STACK_ALLOCATABLE_LISTS
#define LISP_FEATURE_STACK_ALLOCATABLE_VECTORS
#define LISP_FEATURE_STACK_GROWS_DOWNWARD_NOT_UPWARD
#define LISP_FEATURE_SYMBOL_INFO_VOPS
#define LISP_FEATURE_UD2_BREAKPOINTS
#define LISP_FEATURE_UNBIND_N_VOP
#define LISP_FEATURE_UNDEFINED_FUN_RESTARTS
#define LISP_FEATURE_UNIX
#define LISP_FEATURE_UNWIND_TO_FRAME_AND_CALL_VOP
#define LISP_FEATURE_X86_64

#define SBCL_VERSION_STRING "1.4.0"
#define CORE_MAGIC 0x5342434C
#ifndef LANGUAGE_ASSEMBLY

#define LISPOBJ(x) ((lispobj)x)

#else /* LANGUAGE_ASSEMBLY */

#define LISPOBJ(thing) thing

#endif /* LANGUAGE_ASSEMBLY */


#endif
