monitor.o monitor.d : monitor.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h parse.h vars.h print.h arch.h os.h target-os.h \
  darwin-os.h target-arch-os.h target-arch.h interr.h thread.h globals.h \
  runtime-options.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h interrupt.h \
  validate.h search.h purify.h lispregs.h target-lispregs.h \
  genesis/primitive-objects.h genesis/array.h genesis/bignum.h \
  genesis/binding.h genesis/catch-block.h genesis/closure.h \
  genesis/code.h genesis/complex.h genesis/complex-double-float.h \
  genesis/complex-single-float.h genesis/cons.h genesis/double-float.h \
  genesis/funcallable-instance.h genesis/instance.h genesis/ratio.h \
  genesis/sap.h genesis/simd-pack.h genesis/simple-fun.h \
  genesis/unwind-block.h genesis/value-cell.h genesis/vector.h \
  genesis/weak-pointer.h
