darwin-os.o darwin-os.d : darwin-os.c thread.h sbcl.h genesis/config.h \
  genesis/constants.h globals.h runtime.h fixnump.h runtime-options.h \
  os.h target-os.h darwin-os.h target-arch-os.h target-arch.h interr.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h interrupt.h validate.h
