breakpoint.o breakpoint.d : breakpoint.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h os.h target-os.h darwin-os.h target-arch-os.h \
  target-arch.h interr.h interrupt.h thread.h globals.h \
  runtime-options.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h validate.h \
  arch.h lispregs.h target-lispregs.h alloc.h gc-internal.h \
  genesis/code.h genesis/simple-fun.h genesis/weak-pointer.h \
  gencgc-internal.h hopscotch.h genesis/bignum.h genesis/sap.h \
  breakpoint.h
