thread.o thread.d : thread.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h validate.h thread.h globals.h runtime-options.h os.h \
  target-os.h darwin-os.h target-arch-os.h target-arch.h interr.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h interrupt.h arch.h dynbind.h \
  genesis/cons.h alloc.h gc-internal.h genesis/code.h \
  genesis/simple-fun.h genesis/weak-pointer.h gencgc-internal.h \
  hopscotch.h genesis/bignum.h genesis/sap.h cpputil.h pseudo-atomic.h \
  lispregs.h target-lispregs.h genesis/thread-init.inc
