parse.o parse.d : parse.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h globals.h runtime-options.h os.h target-os.h darwin-os.h \
  target-arch-os.h target-arch.h interr.h vars.h parse.h interrupt.h \
  thread.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h validate.h \
  lispregs.h target-lispregs.h monitor.h arch.h search.h pseudo-atomic.h \
  genesis/simple-fun.h
