fullcgc.o fullcgc.d : fullcgc.c gc.h sbcl.h genesis/config.h genesis/constants.h \
  os.h runtime.h fixnump.h target-os.h darwin-os.h target-arch-os.h \
  target-arch.h interr.h gc-internal.h genesis/code.h \
  genesis/simple-fun.h genesis/weak-pointer.h thread.h globals.h \
  runtime-options.h gencgc-alloc-region.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h interrupt.h \
  validate.h gencgc-internal.h hopscotch.h genesis/bignum.h \
  genesis/gc-tables.h genesis/closure.h genesis/layout.h \
  genesis/hash-table.h queue.h weak-hash-pred.inc
