purify.o purify.d : purify.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h os.h target-os.h darwin-os.h target-arch-os.h target-arch.h \
  interr.h globals.h runtime-options.h validate.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h interrupt.h purify.h gc-internal.h \
  genesis/code.h genesis/simple-fun.h genesis/weak-pointer.h \
  gencgc-internal.h hopscotch.h genesis/bignum.h genesis/gc-tables.h \
  genesis/primitive-objects.h genesis/array.h genesis/binding.h \
  genesis/catch-block.h genesis/closure.h genesis/complex.h \
  genesis/complex-double-float.h genesis/complex-single-float.h \
  genesis/cons.h genesis/double-float.h genesis/funcallable-instance.h \
  genesis/instance.h genesis/ratio.h genesis/sap.h genesis/simd-pack.h \
  genesis/unwind-block.h genesis/value-cell.h genesis/vector.h \
  genesis/layout.h genesis/defstruct-description.h genesis/hash-table.h \
  gencgc.h
