dynbind.o dynbind.d : dynbind.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h globals.h runtime-options.h os.h target-os.h \
  darwin-os.h target-arch-os.h target-arch.h interr.h dynbind.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h interrupt.h validate.h \
  genesis/binding.h
