coreparse.o coreparse.d : coreparse.c sbcl.h genesis/config.h genesis/constants.h os.h \
  runtime.h fixnump.h target-os.h darwin-os.h target-arch-os.h \
  target-arch.h interr.h globals.h runtime-options.h core.h arch.h \
  thread.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h interrupt.h \
  validate.h gc-internal.h genesis/code.h genesis/simple-fun.h \
  genesis/weak-pointer.h gencgc-internal.h hopscotch.h genesis/bignum.h \
  pseudo-atomic.h ../../output/build-id.inc genesis/gc-tables.h \
  genesis/hash-table.h genesis/layout.h genesis/vector.h var-io.h
