
/* This is an automatically generated file, please do not hand-edit it.
 * See the program tools-for-build/ldso-stubs.lisp. */

#ifndef LANGUAGE_ASSEMBLY
#define LANGUAGE_ASSEMBLY
#endif
#include "sbcl.h"

#define LDSO_STUBIFY(fct)                       \
        .align 4 ;                              \
.globl _ldso_stub__ ## fct ;                    \
_ldso_stub__ ## fct: ;                          \
        jmp _ ## fct ;                          \
.L ## fct ## e1: ;                            
LDSO_STUBIFY(_exit)
LDSO_STUBIFY(accept)
LDSO_STUBIFY(access)
LDSO_STUBIFY(acos)
LDSO_STUBIFY(acosh)
LDSO_STUBIFY(asin)
LDSO_STUBIFY(asinh)
LDSO_STUBIFY(atanh)
LDSO_STUBIFY(bind)
LDSO_STUBIFY(chmod)
LDSO_STUBIFY(chown)
LDSO_STUBIFY(close)
LDSO_STUBIFY(closedir)
LDSO_STUBIFY(connect)
LDSO_STUBIFY(cosh)
LDSO_STUBIFY(creat)
LDSO_STUBIFY(dup)
LDSO_STUBIFY(dup2)
LDSO_STUBIFY(execve)
LDSO_STUBIFY(exit)
LDSO_STUBIFY(fchmod)
LDSO_STUBIFY(fchown)
LDSO_STUBIFY(fcntl)
LDSO_STUBIFY(fork)
LDSO_STUBIFY(free)
LDSO_STUBIFY(fstat)
LDSO_STUBIFY(fstat$INODE64)
LDSO_STUBIFY(fsync)
LDSO_STUBIFY(ftruncate)
LDSO_STUBIFY(getcwd)
LDSO_STUBIFY(getdtablesize)
LDSO_STUBIFY(getegid)
LDSO_STUBIFY(getenv)
LDSO_STUBIFY(getgid)
LDSO_STUBIFY(gethostbyaddr)
LDSO_STUBIFY(gethostbyname)
LDSO_STUBIFY(gethostname)
LDSO_STUBIFY(getitimer)
LDSO_STUBIFY(getpeername)
LDSO_STUBIFY(getpgrp)
LDSO_STUBIFY(getpid)
LDSO_STUBIFY(getppid)
LDSO_STUBIFY(getrusage)
LDSO_STUBIFY(getsockname)
LDSO_STUBIFY(gettimeofday)
LDSO_STUBIFY(getuid)
LDSO_STUBIFY(hypot)
LDSO_STUBIFY(ioctl)
LDSO_STUBIFY(isatty)
LDSO_STUBIFY(kill)
LDSO_STUBIFY(killpg)
LDSO_STUBIFY(link)
LDSO_STUBIFY(listen)
LDSO_STUBIFY(log1p)
LDSO_STUBIFY(lseek)
LDSO_STUBIFY(lstat)
LDSO_STUBIFY(lstat$INODE64)
LDSO_STUBIFY(malloc)
LDSO_STUBIFY(memcmp)
LDSO_STUBIFY(memmove)
LDSO_STUBIFY(mkdir)
LDSO_STUBIFY(nanosleep)
LDSO_STUBIFY(open)
LDSO_STUBIFY(opendir)
LDSO_STUBIFY(pipe)
LDSO_STUBIFY(poll)
LDSO_STUBIFY(pow)
LDSO_STUBIFY(read)
LDSO_STUBIFY(readdir)
LDSO_STUBIFY(readlink)
LDSO_STUBIFY(realpath)
LDSO_STUBIFY(recv)
LDSO_STUBIFY(rename)
LDSO_STUBIFY(rmdir)
LDSO_STUBIFY(select)
LDSO_STUBIFY(send)
LDSO_STUBIFY(setitimer)
LDSO_STUBIFY(setpgrp)
LDSO_STUBIFY(setsid)
LDSO_STUBIFY(sinh)
LDSO_STUBIFY(socket)
LDSO_STUBIFY(stat)
LDSO_STUBIFY(stat$INODE64)
LDSO_STUBIFY(strerror)
LDSO_STUBIFY(strlen)
LDSO_STUBIFY(symlink)
LDSO_STUBIFY(sync)
LDSO_STUBIFY(tanh)
LDSO_STUBIFY(truncate)
LDSO_STUBIFY(ttyname)
LDSO_STUBIFY(tzname)
LDSO_STUBIFY(unlink)
LDSO_STUBIFY(utimes)
LDSO_STUBIFY(wait3)
LDSO_STUBIFY(waitpid)
LDSO_STUBIFY(write)
LDSO_STUBIFY(sin)
LDSO_STUBIFY(cos)
LDSO_STUBIFY(tan)
LDSO_STUBIFY(atan)
LDSO_STUBIFY(atan2)
LDSO_STUBIFY(exp)
LDSO_STUBIFY(log)
LDSO_STUBIFY(log10)
LDSO_STUBIFY(sqrt)
LDSO_STUBIFY(dlclose)
LDSO_STUBIFY(dlerror)
LDSO_STUBIFY(dlopen)
LDSO_STUBIFY(dlsym)
LDSO_STUBIFY(sysctl)
LDSO_STUBIFY(sysctlbyname)
LDSO_STUBIFY(dladdr)
LDSO_STUBIFY(nl_langinfo)
LDSO_STUBIFY(getpagesize)
LDSO_STUBIFY(cfgetispeed)
LDSO_STUBIFY(cfgetospeed)
LDSO_STUBIFY(cfsetispeed)
LDSO_STUBIFY(cfsetospeed)
LDSO_STUBIFY(tcdrain)
LDSO_STUBIFY(tcflow)
LDSO_STUBIFY(tcflush)
LDSO_STUBIFY(tcgetattr)
LDSO_STUBIFY(tcsendbreak)
LDSO_STUBIFY(tcsetattr)

#ifdef __ELF__
// Mark the object as not requiring an executable stack.
.section .note.GNU-stack,"",%progbits
#endif
