/*
 * This is a machine-generated file. Please do not edit it by hand.
 * (As of sbcl-0.8.14, it came from WRITE-CONFIG-H in genesis.lisp.)
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_PRIMITIVE_OBJECTS
#define SBCL_GENESIS_PRIMITIVE_OBJECTS
#include "array.h"
#include "bignum.h"
#include "binding.h"
#include "catch-block.h"
#include "closure.h"
#include "code.h"
#include "complex.h"
#include "complex-double-float.h"
#include "complex-single-float.h"
#include "cons.h"
#include "double-float.h"
#include "fdefn.h"
#include "funcallable-instance.h"
#include "instance.h"
#include "ratio.h"
#include "sap.h"
#include "simd-pack.h"
#include "simple-fun.h"
#include "symbol.h"
#include "thread.h"
#include "unwind-block.h"
#include "value-cell.h"
#include "vector.h"
#include "weak-pointer.h"
#endif
