x86-64-bsd-os.o x86-64-bsd-os.d : x86-64-bsd-os.c sbcl.h genesis/config.h \
  genesis/constants.h runtime.h fixnump.h thread.h globals.h \
  runtime-options.h os.h target-os.h darwin-os.h target-arch-os.h \
  target-arch.h interr.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h interrupt.h \
  validate.h lispregs.h target-lispregs.h
