/*
 * This is a machine-generated file. Please do not edit it by hand.
 * (As of sbcl-0.8.14, it came from WRITE-CONFIG-H in genesis.lisp.)
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
write_TLS(FREE_INTERRUPT_CONTEXT_INDEX, 0, th);
write_TLS(ALLOW_WITH_INTERRUPTS, T, th);
write_TLS(INTERRUPTS_ENABLED, T, th);
write_TLS(ALLOC_SIGNAL, NIL, th);
write_TLS(INTERRUPT_PENDING, NIL, th);
write_TLS(IN_WITHOUT_GCING, NIL, th);
write_TLS(GC_INHIBIT, NIL, th);
write_TLS(GC_PENDING, NIL, th);
write_TLS(STOP_FOR_GC_PENDING, NIL, th);
write_TLS(CURRENT_CATCH_BLOCK, 0, th);
write_TLS(CURRENT_UNWIND_PROTECT_BLOCK, 0, th);
