save.o save.d : save.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h os.h target-os.h darwin-os.h target-arch-os.h target-arch.h \
  interr.h core.h globals.h runtime-options.h save.h dynbind.h \
  lispregs.h target-lispregs.h validate.h thread.h gencgc-alloc-region.h \
  gc.h genesis/symbol.h genesis/static-symbols.h genesis/thread.h \
  genesis/fdefn.h interrupt.h gc-internal.h genesis/code.h \
  genesis/simple-fun.h genesis/weak-pointer.h gencgc-internal.h \
  hopscotch.h genesis/bignum.h arch.h pseudo-atomic.h
