gc-common.o gc-common.d : gc-common.c sbcl.h genesis/config.h genesis/constants.h \
  runtime.h fixnump.h os.h target-os.h darwin-os.h target-arch-os.h \
  target-arch.h interr.h globals.h runtime-options.h interrupt.h \
  thread.h gencgc-alloc-region.h gc.h genesis/symbol.h \
  genesis/static-symbols.h genesis/thread.h genesis/fdefn.h validate.h \
  lispregs.h target-lispregs.h arch.h hopscotch.h \
  genesis/primitive-objects.h genesis/array.h genesis/bignum.h \
  genesis/binding.h genesis/catch-block.h genesis/closure.h \
  genesis/code.h genesis/complex.h genesis/complex-double-float.h \
  genesis/complex-single-float.h genesis/cons.h genesis/double-float.h \
  genesis/funcallable-instance.h genesis/instance.h genesis/ratio.h \
  genesis/sap.h genesis/simd-pack.h genesis/simple-fun.h \
  genesis/unwind-block.h genesis/value-cell.h genesis/vector.h \
  genesis/weak-pointer.h genesis/layout.h genesis/hash-table.h \
  gc-internal.h gencgc-internal.h forwarding-ptr.h var-io.h \
  weak-hash-pred.inc genesis/gc-tables.h
