print.o print.d : print.c sbcl.h genesis/config.h genesis/constants.h print.h \
  runtime.h fixnump.h gc-internal.h genesis/code.h genesis/simple-fun.h \
  genesis/weak-pointer.h thread.h globals.h runtime-options.h os.h \
  target-os.h darwin-os.h target-arch-os.h target-arch.h interr.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h interrupt.h validate.h \
  gencgc-internal.h hopscotch.h genesis/bignum.h monitor.h vars.h \
  genesis/primitive-objects.h genesis/array.h genesis/binding.h \
  genesis/catch-block.h genesis/closure.h genesis/complex.h \
  genesis/complex-double-float.h genesis/complex-single-float.h \
  genesis/cons.h genesis/double-float.h genesis/funcallable-instance.h \
  genesis/instance.h genesis/ratio.h genesis/sap.h genesis/simd-pack.h \
  genesis/unwind-block.h genesis/value-cell.h genesis/vector.h \
  genesis/tagnames.h genesis/layout.h forwarding-ptr.h \
  genesis/classoid.h
