gencgc.o gencgc.d : gencgc.c sbcl.h genesis/config.h genesis/constants.h runtime.h \
  fixnump.h os.h target-os.h darwin-os.h target-arch-os.h target-arch.h \
  interr.h globals.h runtime-options.h interrupt.h thread.h \
  gencgc-alloc-region.h gc.h genesis/symbol.h genesis/static-symbols.h \
  genesis/thread.h genesis/fdefn.h validate.h lispregs.h \
  target-lispregs.h arch.h gc-internal.h genesis/code.h \
  genesis/simple-fun.h genesis/weak-pointer.h gencgc-internal.h \
  hopscotch.h genesis/bignum.h pseudo-atomic.h alloc.h genesis/sap.h \
  genesis/gc-tables.h genesis/vector.h save.h core.h \
  genesis/hash-table.h genesis/instance.h genesis/layout.h gencgc.h \
  genesis/cons.h forwarding-ptr.h
