os-common.o os-common.d : os-common.c sbcl.h genesis/config.h genesis/constants.h \
  globals.h runtime.h fixnump.h runtime-options.h os.h target-os.h \
  darwin-os.h target-arch-os.h target-arch.h interr.h genesis/cons.h \
  genesis/vector.h genesis/symbol.h genesis/static-symbols.h thread.h \
  gencgc-alloc-region.h gc.h genesis/thread.h genesis/fdefn.h \
  interrupt.h validate.h arch.h
